//
//  Copyright (c) 2022 Open Whisper Systems. All rights reserved.
//

import Foundation
import SignalUI
import UIKit

enum BadgeConfiguration {
    case doNotDisplayPublicly
    case display(featuredBadge: OWSUserProfileBadgeInfo)
}

protocol BadgeConfigurationDelegate: AnyObject {
    func badgeConfiguration(_: BadgeConfigurationViewController, didCompleteWithBadgeSetting setting: BadgeConfiguration)
    func badgeConfirmationDidCancel(_: BadgeConfigurationViewController)
}

class BadgeConfigurationViewController: OWSTableViewController2, BadgeCollectionDataSource, OWSNavigationView {
    private weak var badgeConfigDelegate: BadgeConfigurationDelegate?

    let availableBadges: [OWSUserProfileBadgeInfo]
    private let initialDisplaySetting: Bool
    private let avatarImage: UIImage?

    private var displayBadgeOnProfile: Bool {
        didSet {
            updateNavigation()
            updateTableContents()
        }
    }

    var selectedBadgeIndex: Int = 0 {
        didSet {
            if !availableBadges.indices.contains(selectedBadgeIndex) {
                owsFailDebug("Invalid badge index")
                selectedBadgeIndex = oldValue
            } else {
                updateNavigation()
                updateTableContents()
            }
        }
    }

    private var selectedPrimaryBadge: OWSUserProfileBadgeInfo? {
        displayBadgeOnProfile ? availableBadges[safe: selectedBadgeIndex] : nil
    }

    private var hasUnsavedChanges: Bool {
        if displayBadgeOnProfile, selectedBadgeIndex != 0 {
            return true
        } else {
            return displayBadgeOnProfile != initialDisplaySetting
        }
    }

    public var showDismissalActivity = false {
        didSet {
            updateNavigation()
        }
    }

    convenience init(fetchingDataFromLocalProfileWithDelegate delegate: BadgeConfigurationDelegate) {
        let snapshot = Self.profileManagerImpl.localProfileSnapshot(shouldIncludeAvatar: false)
        let allBadges = snapshot.profileBadgeInfo ?? []
        let shouldDisplayOnProfile = Self.subscriptionManager.displayBadgesOnProfile

        self.init(availableBadges: allBadges, shouldDisplayOnProfile: shouldDisplayOnProfile, delegate: delegate)
    }

    init(availableBadges: [OWSUserProfileBadgeInfo], shouldDisplayOnProfile: Bool, avatarImage: UIImage? = nil, delegate: BadgeConfigurationDelegate) {
        self.availableBadges = availableBadges
        self.initialDisplaySetting = shouldDisplayOnProfile
        owsAssertDebug(availableBadges.indices.contains(selectedBadgeIndex))

        self.displayBadgeOnProfile = self.initialDisplaySetting
        self.badgeConfigDelegate = delegate
        self.avatarImage = avatarImage
        super.init()
    }

    override func viewDidLoad() {
        super.viewDidLoad()
        updateNavigation()
        updateTableContents()
    }

    // MARK: - Navigation Bar

    private func updateNavigation() {
        if navigationController?.viewControllers.count == 1 {
            navigationItem.leftBarButtonItem = UIBarButtonItem(
                barButtonSystemItem: .cancel,
                target: self,
                action: #selector(didTapCancel),
                accessibilityIdentifier: "cancel_button")
        }

        if hasUnsavedChanges, showDismissalActivity {
            let indicatorStyle: UIActivityIndicatorView.Style
            if #available(iOS 13, *) {
                indicatorStyle = .medium
            } else {
                indicatorStyle = Theme.isDarkThemeEnabled ? .white : .gray
            }
            let spinner = UIActivityIndicatorView(style: indicatorStyle)
            spinner.startAnimating()
            navigationItem.rightBarButtonItem = UIBarButtonItem(customView: spinner)
        } else if hasUnsavedChanges, !showDismissalActivity {
            navigationItem.rightBarButtonItem = UIBarButtonItem(
                title: CommonStrings.setButton,
                style: .done,
                target: self,
                action: #selector(didTapDone),
                accessibilityIdentifier: "set_button")
        } else {
            navigationItem.rightBarButtonItem = nil
        }
    }

    @objc
    func didTapCancel() {
        let requestDismissal: () -> Void = { [weak self] in
            guard let self = self else { return }
            self.badgeConfigDelegate?.badgeConfirmationDidCancel(self)
        }

        if hasUnsavedChanges {
            OWSActionSheets.showPendingChangesActionSheet(discardAction: requestDismissal)
        } else {
            requestDismissal()
        }
    }

    @objc
    func didTapDone() {
        if displayBadgeOnProfile, let selectedPrimaryBadge = selectedPrimaryBadge {
            badgeConfigDelegate?.badgeConfiguration(self, didCompleteWithBadgeSetting: .display(featuredBadge: selectedPrimaryBadge))
        } else {
            badgeConfigDelegate?.badgeConfiguration(self, didCompleteWithBadgeSetting: .doNotDisplayPublicly)
        }
    }

    override var isModalInPresentation: Bool {
        get { hasUnsavedChanges }
        set { owsFailDebug("Unsupported") }
    }

    // MARK: - TableContents

    func updateTableContents() {
        self.contents = OWSTableContents(
            title: NSLocalizedString("BADGE_CONFIGURATION_TITLE", comment: "The title for the badge configuration page"),
            sections: [
                OWSTableSection(
                    title: NSLocalizedString(
                        "BADGE_CONFIGURATION_BADGE_SECTION_TITLE",
                        comment: "Section header for badge view section in the badge configuration page"),
                    items: [
                        OWSTableItem(customCellBlock: { [weak self] in
                            let cell = OWSTableItem.newCell()
                            guard let self = self else { return cell }
                            let collectionView = BadgeCollectionView(dataSource: self)

                            if let localAddress = self.tsAccountManager.localAddress {
                                let localShortName = self.databaseStorage.read { self.contactsManager.shortDisplayName(for: localAddress, transaction: $0) }
                                collectionView.badgeSelectionMode = .detailsSheet(owner: .local(shortName: localShortName))
                            } else {
                                owsFailDebug("Unexpectedly missing local address")
                            }

                            cell.contentView.addSubview(collectionView)
                            collectionView.autoPinEdgesToSuperviewMargins()

                            // Pre-layout the collection view so the UITableView caches the correct resolved
                            // autolayout height.
                            collectionView.layoutIfNeeded()

                            return cell
                        }, actionBlock: nil)
                ]),

                OWSTableSection(title: nil, items: [
                    .switch(
                        withText: NSLocalizedString(
                            "DISPLAY_BADGES_ON_PROFILE_SETTING",
                            comment: "Title for switch to enable sharing of badges publicly"),
                        isOn: { [weak self] in self?.displayBadgeOnProfile ?? false },
                        target: self,
                        selector: #selector(didTogglePublicDisplaySetting(_:))),

                    .item(
                        name: NSLocalizedString(
                            "FEATURED_BADGE_SETTINGS_TITLE",
                            comment: "The title for the featured badge settings page"),
                        textColor: displayBadgeOnProfile ? nil : .ows_gray45,
                        accessoryText: selectedPrimaryBadge?.badge?.localizedName,
                        accessoryType: .disclosureIndicator,
                        accessibilityIdentifier: "badge_configuration_row",
                        actionBlock: { [weak self] in
                            guard let self = self, let navController = self.navigationController else { return }
                            guard self.displayBadgeOnProfile else { return }

                            let featuredBadgeSettings = FeaturedBadgeViewController(avatarImage: self.avatarImage, badgeDataSource: self)
                            navController.pushViewController(featuredBadgeSettings, animated: true)
                        })
                ])
        ])
    }

    @objc
    func didTogglePublicDisplaySetting(_ sender: UISwitch) {
        displayBadgeOnProfile = sender.isOn
    }

    func shouldCancelNavigationBack() -> Bool {
        if hasUnsavedChanges {
            didTapCancel()
            return true
        } else {
            return false
        }
    }
}
